# -*- coding=UTF-8 -*-
"""Apollo-like resolver.  """

from __future__ import annotations
import typing

import graphene
import graphql

from . import schema as schema_


class Resolver:
    """Apollo-like schema field resolver.  """

    # Resolver definitions.
    schema: typing.Optional[typing.MutableMapping] = None

    # Data that avaliable inside `resolve`.
    parent: typing.Any
    info: graphql.execution.base.ResolveInfo
    context: typing.Any

    _field: typing.Optional[graphene.Field] = None
    _schema: schema_.FieldDefinition
    _type: typing.Optional[graphene.Scalar | graphene.ObjectType] = None
    _as_interface: typing.Optional[typing.Type[graphene.Interface]] = None

    def __init_subclass__(cls):
        cls._parse_schema(default={'name': cls.__name__})
        cls.as_type()

    def __init__(
            self,
            *,
            info: graphql.execution.base.ResolveInfo,
            parent: typing.Any = None,
    ):
        self.parent = parent
        self.info = info
        self.context = info.context

    def resolve(self, **kwargs):
        """Resolve the field.  """
        # pylint:disable=unused-argument

        field_name = self.info.field_name
        if isinstance(self.parent, typing.Mapping) and field_name in self.parent:
            return self.parent[field_name]
        elif hasattr(self.parent, field_name):
            return getattr(self.parent, field_name)
        raise NotImplementedError(
            f'`{self.__class__.__name__}.resolve` is not implemented.')

    def get_node(self, id_: str):
        """Get node value from id.

        Args:
            id_ (str): Node id.

        Returns:
            typing.Any: Corresponding node value.
        """
        # pylint:disable=unused-argument,no-self-use
        return None

    def validate(self, value) -> bool:
        """Test whether value is match resolver schema type.

        Args:
            value (typing.Any): Value to validate.

        Returns:
            bool: whether value match schema type.
        """
        # pylint:disable=unused-argument,no-self-use
        return True

    @classmethod
    def _parse_schema(cls, *, default: typing.Dict):
        if not cls.schema:
            raise NotImplementedError(
                f'Resolver schema is not defined: {cls.__name__}')

        def resolve_fn(parent, info: graphql.execution.base.ResolveInfo, **kwargs):
            ret = cls(parent=parent, info=info).resolve(**kwargs)
            if isinstance(ret, typing.Mapping) and '__typename' in ret:
                type_ = info.schema.get_type(ret['__typename']).graphene_type
                ret = type_(
                    **{k: v for k, v in ret.items()
                       if k in type_._meta.fields})
            return ret

        cls._schema = schema_.FieldDefinition.parse(
            cls.schema,
            default={**default, 'resolver': resolve_fn}
        )
        return cls._schema

    @classmethod
    def as_type(
            cls,
    ) -> graphene.types.unmountedtype.UnmountedType:
        """Convert resolver as graphene type.

        Args:
            namespace (str, optional): [description]. Defaults to None.
                namespace for auto naming, use class name when value is None.

        Returns:
            graphene.types.unmountedtype.UnmountedType:
        """

        if cls._type:
            return cls._type

        ret = cls._schema.as_type()

        def get_node(info, id_):
            return cls(info=info).get_node(id_)
        ret.get_node = get_node

        def is_type_of(value, info):
            return cls(info=info).validate(value)
        ret.is_type_of = is_type_of

        cls._type = ret
        return ret

    @classmethod
    def as_field(cls) -> graphene.Field:
        """Convert resolver as a graphene field.

        Returns:
            graphene.Field: Converted field.
        """

        if cls._field:
            return cls._field

        cls._field = cls._schema.mount(type_=cls.as_type(), as_=graphene.Field)
        return cls._field

    @classmethod
    def as_interface(cls) -> typing.Type[graphene.Interface]:
        """Convert resolver schema as interface

        Raises:
            ValueError: Can not convert to interface.

        Returns:
            typing.Type[graphene.Interface]: Convert result, will cache on class.
        """
        if cls._schema.type is not schema_.SpecialType.MAPPING:
            raise ValueError(
                f'Schema can not use as interface, should be mapping: {cls.__name__}')

        if cls._as_interface:
            return cls._as_interface
        cls._as_interface = cls._schema.as_type(
            mapping_bases=(graphene.Interface,))
        return cls._as_interface



"""Graphene scalar types.  """

from datetime import timedelta

import graphene
import isodate
from graphql.language import ast


class Duration(graphene.Scalar):
    """Duration in ISO-8601 format.  """

    @staticmethod
    def serialize(duration: timedelta):
        """Serialize python object.

        Args:
            duration (timedelta): Duration

        Returns:
            str
        """

        return isodate.duration_isoformat(duration)

    @classmethod
    def parse_literal(cls, node):
        """Parse ast node.

        Args:
            node: AST node

        Returns:
            timedelta | None
        """

        if isinstance(node, ast.StringValue):
            return cls.parse_value(node.value)
        return None

    @staticmethod
    def parse_value(value):
        """Parse str to python object.

        Args:
            value (str): Value

        Returns:
            timedelta | None
        """

        try:
            return isodate.parse_duration(value)
        except ValueError:
            return None



# -*- coding=UTF-8 -*-
"""Mongoose-like schema.  """

from __future__ import annotations
import functools
import typing

import graphene

from . import types

REGISTRY: typing.Dict[
    str,
    typing.Type[graphene.types.unmountedtype.UnmountedType]
] = {
    'ID': graphene.ID,
    'Boolean': graphene.Boolean,
    'String': graphene.String,
    'Int': graphene.Int,
    'Float': graphene.Float,
    'Decimal': graphene.Decimal,
    'Time': graphene.Time,
    'Date': graphene.Date,
    'DateTime': graphene.DateTime,
    'Duration': types.Duration,
}


def dynamic_type(type_: typing.Any, *, registry=None) -> typing.Callable:
    """Get dynamic type function for given typename.

    Args:
        type_ (typing.Any): typename or type itself.
        registry (typing.Dict, optional): Graphene type registry. Defaults to None.

    Returns:
        typing.Callable: dynamic type function
    """
    registry = registry or REGISTRY

    def type_fn(type_, *_args, **_kwargs):
        if isinstance(type_, str):
            type_ = registry[type_]

        assert (isinstance(type_, type)), repr(type_)
        return type_

    return functools.partial(type_fn, type_)



# -*- coding=UTF-8 -*-
"""Mongoose-like schema.  """
# pylint:disable=unused-import

from __future__ import annotations
import dataclasses
import enum
import typing

import graphene
import phrases_case

from . import typedef
from . import processor


class SpecialType(enum.Enum):
    """Indicate special type syntax used in schema.  """

    MAPPING = enum.auto()
    LIST = enum.auto()
    ENUM = enum.auto()
    UNION = enum.auto()


CONFIG_PROCESSOR = processor.Processor()


@CONFIG_PROCESSOR.register(100)
def _process_resolver_config(type_def, config):
    from . import resolver
    if not (isinstance(type_def, type)
            and issubclass(type_def, resolver.Resolver)):
        return None
    _resolver: typing.Type[resolver.Resolver] = type_def
    # merge schema
    config['name'] = _resolver._schema.name
    config['type'] = _resolver.as_type()
    config['args'].update(**_resolver._schema.args)
    config['interfaces'] += _resolver._schema.interfaces
    config.setdefault('required', _resolver._schema.required)

    _parent_resolver = config.get('resolver')

    def resolve_fn(parent, info, **kwargs):
        if _parent_resolver:
            parent = _parent_resolver(parent, info, **kwargs)
        return _resolver._schema.resolver(parent, info, **kwargs)
    config.setdefault('resolver', resolve_fn)
    config.setdefault('description', _resolver._schema.description)
    config.setdefault('deprecation_reason',
                      _resolver._schema.deprecation_reason)
    child_definition = _resolver._schema.child_definition
    return dict(
        config=config,
        child_definition=child_definition
    )


@CONFIG_PROCESSOR.register(90)
def _process_str_type_def(type_def, config):
    if not isinstance(type_def, str):
        return None
    if type_def[-1] == '!':
        config.setdefault('required', True)
        type_def = type_def[:-1]
    config['type'] = type_def
    return dict(
        config=config
    )


@CONFIG_PROCESSOR.register(80)
def _process_mapping_type_def(type_def, config):
    if not isinstance(type_def, typing.Mapping):
        return None
    config['type'] = SpecialType.MAPPING
    return dict(
        config=config,
        child_definition=type_def
    )


@CONFIG_PROCESSOR.register(70)
def _process_empty_iterable_type_def(type_def, config):
    if isinstance(type_def, typing.Iterable) and len(type_def) == 0:
        raise ValueError('Can not use empty iterable as type.')


@CONFIG_PROCESSOR.register(60)
def _process_list_type_def(type_def, config):
    if not (isinstance(type_def, typing.Iterable) and len(type_def) == 1):
        return None
    config['type'] = SpecialType.LIST
    return dict(
        config=config,
        child_definition=type_def[0]
    )


@CONFIG_PROCESSOR.register(50)
def _process_enum_type_def(type_def, config):
    if not(isinstance(type_def, typing.Iterable)
            and all(isinstance(i, (str, tuple)) for i in type_def)):
        return None
    config['type'] = SpecialType.ENUM
    return dict(
        config=config,
        child_definition=type_def
    )


@CONFIG_PROCESSOR.register(40)
def _process_union_type_def(type_def, config):
    if not isinstance(type_def, typing.Iterable):
        return None
    config['type'] = SpecialType.UNION
    return dict(
        config=config,
        child_definition=type_def
    )


@CONFIG_PROCESSOR.register(-1)
def _process_type_type_def(type_def, config):
    config['type'] = type_def
    return dict(
        config=config,
    )


@dataclasses.dataclass
class EnumFieldDefinition:
    value: str
    description: typing.Optional[str]

    @classmethod
    def parse(cls, v) -> EnumFieldDefinition:
        """Parse schema for enum

        Args:
            v (typing.Any): schema item

        Returns:
            EnumFieldDefinition: Parsing result
        """
        if isinstance(v, str):
            return cls(value=v, description=None)
        elif isinstance(v, tuple) and len(v) == 2:
            return cls(value=v[0], description=v[1])
        raise ValueError(
            f'Enum field should be a str or 2-value tuple, got {v}')


@dataclasses.dataclass
class FieldDefinition:
    """A mongoose-like schema for resolver field.  """

    # Options:
    args: typing.Mapping
    type: typing.Type
    required: bool
    name: str
    interfaces: typing.Tuple[typing.Type[graphene.Interface], ...]
    description: typing.Optional[str]
    deprecation_reason: typing.Optional[str]
    resolver: typing.Optional[typing.Callable]
    default: typing.Any

    # Parse results:
    child_definition: typing.Any

    @classmethod
    def parse(cls, v: typing.Any, *, default: typing.Dict = None) -> FieldDefinition:
        """Parse a mongoose like schema

        Args:
            v (typing.Any): schema item

        Returns:
            SchemaDefinition: Parsing result
        """
        from . import resolver

        assert v is not None, 'schema is None'
        config = default or {}
        child_definition = None
        is_full_config = (
            isinstance(v, typing.Mapping)
            and 'type' in v
            and not (  # Handle `type` as mapping key.
                isinstance(v['type'], typing.Mapping)
                and 'type' in v['type']
                and not isinstance(v['type']['type'], typing.Mapping)))

        type_def = v
        if is_full_config:
            config.update(**v)
            type_def = v['type']
        if 'name' not in config:
            raise ValueError(f'Not specified field name in: {v}')
        # Convert args
        config.setdefault('args', {})
        if config['args']:
            config['args'] = {
                k: (cls
                    .parse(
                        v,
                        default={
                            'name': phrases_case.camel(f'{config["name"]}_{k}')
                        })
                    .mount(as_=graphene.Argument))
                for k, v in config['args'].items()
            }
        # Convert interfaces
        config.setdefault('interfaces', ())
        config['interfaces'] = tuple(
            i.as_interface() if (isinstance(i, type) and issubclass(i, resolver.Resolver)) else i
            for i in config['interfaces']
        )

        result = CONFIG_PROCESSOR.process(config=config, type_def=type_def)
        config = result['config']
        child_definition = result.get('child_definition')

        config.setdefault('required', False)
        config.setdefault('description', None)
        config.setdefault('deprecation_reason', None)
        config.setdefault('resolver', None)
        config.setdefault('default', None)

        return cls(
            type=config['type'],
            required=config['required'],
            name=config['name'],
            description=config['description'],
            deprecation_reason=config['deprecation_reason'],
            args=config['args'],
            interfaces=config['interfaces'],
            resolver=config['resolver'],
            default=config['default'],
            child_definition=child_definition,
        )

    def _get_options(self, target: typing.Type):
        allowed_options = ()
        if not isinstance(target, type):
            pass

        allowed_options_by_type = [
            (graphene.Argument,
             ('required', 'default_value', 'description', )),
            (graphene.Field,
             ('args', 'resolver', 'required', 'default_value',
              'description', 'deprecation_reason', )),
            (graphene.InputField,
             ('required', 'default_value', 'description', 'deprecation_reason',)),
        ]
        for classinfo, options in allowed_options_by_type:
            if issubclass(target, classinfo):
                allowed_options += options

        options = dict(
            name=self.name,
            required=self.required,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            args=self.args,
            resolver=self.resolver,
            default_value=self.default,
        )

        return {k: v for k, v in options.items() if k in allowed_options}

    def as_type(
            self,
            *,
            mapping_bases: typing.Tuple[typing.Type] = (graphene.ObjectType,),
            registry=None
    ) -> graphene.types.unmountedtype.UnmountedType:
        """Convert schema to graphene unmounted type instance with options set.

        Args:
            is_input (bool, optional): Whether is input field. Defaults to False.
            registry (typing.Dict, optional): Graphene type registry. Defaults to None.

        Returns:
            graphene.types.unmountedtype.UnmountedType: result
        """
        registry = registry or typedef.REGISTRY

        namespace = self.name
        is_input = graphene.InputObjectType in mapping_bases

        options = self._get_options(
            graphene.Argument if is_input else graphene.Field)
        ret = None
        # Mapping
        if self.type is SpecialType.MAPPING:
            assert self.child_definition
            _type: typing.Type = type(
                namespace,
                mapping_bases,
                {
                    **{
                        k: (self.parse(
                            v,
                            default={
                                'name': phrases_case.camel(f'{namespace}_{k}')}
                        ).mount(as_=graphene.InputField if is_input else graphene.Field))
                        for k, v in self.child_definition.items()
                    },
                    **dict(
                        Meta=dict(
                            name=self.name,
                            interfaces=self.interfaces,
                            description=self.description,
                        )
                    )
                })
            registry[namespace] = _type
            ret = _type
        # Iterable
        elif self.type is SpecialType.LIST:
            assert self.child_definition
            _item_schema = self.parse(
                self.child_definition,
                default={'name': namespace}
            )
            _item_type = _item_schema.as_type(mapping_bases=mapping_bases)
            if _item_schema.required:
                # `required` option for list item not work,
                # so non-null structure is required.
                _item_type = graphene.NonNull(_item_type)
            ret = graphene.List(
                _item_type,
                **options
            )
        elif self.type is SpecialType.ENUM:
            assert self.child_definition
            _enum_defs = [EnumFieldDefinition.parse(i)
                          for i in self.child_definition]
            _enum = enum.Enum(  # type: ignore
                namespace, {i.value: i.value for i in _enum_defs})

            def _get_description(v):
                if v is None:
                    return self.description
                return next(i for i in _enum_defs if i.value == v.value).description
            ret = graphene.Enum.from_enum(
                _enum,
                description=_get_description
            )
        elif self.type is SpecialType.UNION:
            assert self.child_definition

            def _dynamic():
                _types = [FieldDefinition.parse(i, default={'name': f'{namespace}{index}'}).as_type()
                          for index, i in enumerate(self.child_definition)]
                _types = [i() if callable(i) and not isinstance(i, type) else i
                          for i in _types]
                return type(namespace, (graphene.Union,), dict(
                    Meta=dict(
                        types=_types,
                        description=self.description,
                    )
                ))
            ret = _dynamic

        # Unmounted type.
        elif (isinstance(self.type, type)
              and issubclass(self.type, graphene.types.unmountedtype.UnmountedType)):
            ret = self.type(**options)
        # Dynamic
        elif isinstance(self.type, str):
            ret = typedef.dynamic_type(self.type)
        # As-is
        else:
            ret = self.type
        return ret

    def mount(
            self,
            *,
            as_: typing.Type,
            type_: graphene.types.unmountedtype.UnmountedType = None,
            registry=None
    ):
        """Mount schema as a graphene mounted type instance.

        Args:
            as_ (typing.Type): Target type
            type_ (graphene.types.unmountedtype.UnmountedType, optional):
                Override unmmounted type. Defaults to None.
            registry (typing.Mapping, optional): Graphene type registry. Defaults to None.

        Returns:
            Mounted type instance.
        """

        is_input = (
            isinstance(as_, type)
            and issubclass(
                as_,
                (graphene.Argument, graphene.InputField, graphene.InputObjectType)))
        mapping_bases = (graphene.InputObjectType,
                         ) if is_input else (graphene.ObjectType,)
        type_ = type_ or self.as_type(
            mapping_bases=mapping_bases, registry=registry)

        if isinstance(type_, graphene.types.unmountedtype.UnmountedType):
            return type_.mount_as(as_)
        return as_(type=type_, **self._get_options(as_))



"""Simple weighted function execution.  """

import typing


class Processor:
    """A processor can register multiple process, they will executed by weight
    and stop on first value returns.  """

    _process_registry: typing.List[typing.Tuple[float, typing.Callable]]

    def __init__(self) -> None:
        self._process_registry = []

    def register(self, weight: float) -> typing.Callable[[typing.Callable], None]:
        """Get decorator for process registering.

        Args:
            weight (float): Function weight, higher weight will executed first.

        Returns:
            typing.Callable[[typing.Callable], None]: Decorator.
        """

        def _decorator(func: typing.Callable) -> None:
            self._process_registry.append((weight, func))
        return _decorator

    def process(self, **kwargs) -> dict:
        """Execute all registered function by weight,
        high weight function execute first.

        Raises:
            NotImplementedError: All function is executed but no one returns a value.

        Returns:
            dict: First return value.
        """
        self._process_registry.sort(key=lambda v: v[0], reverse=True)
        for _, i in self._process_registry:
            ret = i(**kwargs)
            if ret is not None:
                return ret
        raise NotImplementedError('Process not implemented')



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['graphene_resolver', 'graphene_resolver.types']

package_data = \
{'': ['*']}

install_requires = \
['graphene>=2.1,<3.0',
 'isodate>=0.6,<0.7',
 'lazy-object-proxy>=1.4,<2.0',
 'phrases-case>=0.1.1,<0.2.0']

setup_kwargs = {
    'name': 'graphene-resolver',
    'version': '0.1.0',
    'description': 'Using mongoose-like schema to write apollo-like resolver for graphene.',
    'long_description': '# Graphene resolvers\n\n[![build status](https://github.com/NateScarlet/graphene-resolver/workflows/Python%20package/badge.svg)](https://github.com/NateScarlet/graphene-resolver/actions)\n[![version](https://img.shields.io/pypi/v/graphene-resolver)](https://pypi.org/project/graphene-resolver/)\n![python version](https://img.shields.io/pypi/pyversions/graphene-resolver)\n![wheel](https://img.shields.io/pypi/wheel/graphene-resolver)\n![maintenance](https://img.shields.io/maintenance/yes/2019)\n\nUsing mongoose-like schema to write apollo-like resolver.\n\n## Install\n\n`pip install graphene-resolver`\n\n## Usage\n\nsimple example:\n\n```python\nimport graphene\nimport graphene_resolver as resolver\n\nclass Foo(resolver.Resolver):\n    schema = {\n        "args": {\n            "key":  \'String!\',\n            "value": \'String!\',\n        },\n        "type": \'String!\',\n    }\n\n    def resolve(self, **kwargs):\n        self.parent # parent field\n        self.info # resolve info\n        self.context # resolve context\n        return kwargs[\'value\']\n\nclass Query(graphene.ObjectType):\n    foo = Foo.as_field()\n```\n\n```graphql\n{\n  foo(key: "k", value: "v")\n}\n```\n\n```json\n{ "foo": "v" }\n```\n\nrelay node:\n\n```python\npets = [dict(\n    id=1,\n    name=\'pet1\',\n    age=1,\n)]\nclass Pet(resolver.Resolver):\n    schema = {\n        \'type\': {\n            \'name\': \'String\',\n            \'age\': \'Int\',\n        },\n        \'interfaces\': (graphene.Node,)\n    }\n\n    def get_node(self, id_):\n        return next(i for i in pets if i[\'id\'] == int(id_))\n\n    def validate(self, value):\n        return (\n            isinstance(value, typing.Mapping)\n            and isinstance(value.get(\'name\'), str)\n            and isinstance(value.get(\'age\'), int)\n        )\nclass Query(graphene.ObjectType):\n    node = graphene.Node.Field()\n\nschema = graphene.Schema(query=Query, types=[Pet.as_type()])\n```\n\n```graphql\n{\n  node(id: "UGV0OjE=") {\n    id\n    __typename\n    ... on Pet {\n      name\n      age\n    }\n  }\n}\n```\n\n```json\n{ "node": { "id": "UGV0OjE=", "__typename": "Pet", "name": "pet1", "age": 1 } }\n```\n\nrelay connection:\n\n```python\nclass Item(resolver.Resolver):\n    schema = {\'name\': \'String!\'}\n\nclass Items(resolver.Resolver):\n    schema = resolver.connection.get_type(Item)\n\n    def resolve(self, **kwargs):\n        return resolver.connection.resolve([{\'name\': \'a\'}, {\'name\': \'b\'}], **kwargs)\n```\n\n```graphql\n{\n  items {\n    edges {\n      node {\n        name\n      }\n      cursor\n    }\n    pageInfo {\n      total\n      hasNextPage\n      hasPreviousPage\n      startCursor\n      endCursor\n    }\n  }\n}\n```\n\n```json\n{\n  "items": {\n    "edges": [\n      { "node": { "name": "a" }, "cursor": "YXJyYXljb25uZWN0aW9uOjA=" },\n      { "node": { "name": "b" }, "cursor": "YXJyYXljb25uZWN0aW9uOjE=" }\n    ],\n    "pageInfo": {\n      "total": 2,\n      "hasNextPage": false,\n      "hasPreviousPage": false,\n      "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",\n      "endCursor": "YXJyYXljb25uZWN0aW9uOjE="\n    }\n  }\n}\n```\n\nenum:\n\n```python\n\n    class Foo(resolver.Resolver):\n        schema = (\'a\', \'b\')\n\n        def resolve(self, **kwargs):\n            return \'a\'\n\n    class Query(graphene.ObjectType):\n        foo = Foo.as_field()\n\n    schema = graphene.Schema(query=Query)\n    assert str(schema) == \'\'\'\\\nschema {\n  query: Query\n}\n\nenum Foo {\n  a\n  b\n}\n\ntype Query {\n  foo: Foo\n}\n\'\'\'\n```\n\nenum with description:\n\n```python\n\n    class Foo(resolver.Resolver):\n        schema = {\n            \'type\': [(\'a\', \'this is a\'), (\'b\', \'this is b\'), \'c\'],\n            \'description\': \'A enum\',\n        }\n\n        def resolve(self, **kwargs):\n            return \'a\'\n\n    class Query(graphene.ObjectType):\n        foo = Foo.as_field()\n\n    schema = graphene.Schema(query=Query)\n    enum_type = schema.get_type(\'Foo\')\n    assert enum_type.description == \'A enum\'\n    assert enum_type.get_value(\'a\').value == \'a\'\n    assert enum_type.get_value(\'a\').description == \'this is a\'\n    assert enum_type.get_value(\'b\').value == \'b\'\n    assert enum_type.get_value(\'b\').description == \'this is b\'\n    assert enum_type.get_value(\'c\').value == \'c\'\n    assert enum_type.get_value(\'c\').description is None\n```\n\nunion:\n\n```python\n    class Foo(resolver.Resolver):\n        schema = ({\'a\': \'String\'}, {\'b\': \'Int\'})\n\n        def resolve(self, **kwargs):\n            return {\'__typename\': \'Foo0\', \'a\': \'a\'}\n\n    class Query(graphene.ObjectType):\n        foo = Foo.as_field()\n\n    schema = graphene.Schema(query=Query)\n    assert str(schema) == \'\'\'\\\nschema {\n  query: Query\n}\n\nunion Foo = Foo0 | Foo1\n\ntype Foo0 {\n  a: String\n}\n\ntype Foo1 {\n  b: Int\n}\n\ntype Query {\n  foo: Foo\n}\n\'\'\'\n```\n\n```graphql\n{\n  foo {\n    __typename\n    ... on Foo0 {\n      a\n    }\n  }\n}\n```\n\n```json\n{ "foo": { "__typename": "Foo0", "a": "a" } }\n```\n\ncomplicated example:\n\n```python\nclass Foo(resolver.Resolver):\n    _input_schema = {\n        "type": {"type": \'String\'},\n        "data": [\n            {\n                "type":\n                {\n                    "key": {\n                        "type": \'String\',\n                        "required": True,\n                        "description": "<description>",\n                    },\n                    "value": \'Int\',\n                    "extra": {\n                        "type": [\'String!\'],\n                        "deprecation_reason": "<deprecated>"\n                    },\n                },\n                "required": True\n            },\n        ],\n    }\n    schema = {\n        "args": {\n            "input": _input_schema\n        },\n        "type": _input_schema,\n        "description": "description",\n        "deprecation_reason": None\n    }\n\n    def resolve(self, **kwargs):\n        return kwargs[\'input\']\n```\n\n```graphql\n{\n  foo(\n    input: { type: "type", data: [{ key: "key", value: 42, extra: ["extra"] }] }\n  ) {\n    type\n    data {\n      key\n      value\n      extra\n    }\n  }\n}\n```\n\n```json\n{\n  "foo": {\n    "type": "type",\n    "data": [{ "key": "key", "value": 42, "extra": ["extra"] }]\n  }\n}\n```\n',
    'author': 'NateScarlet',
    'author_email': 'NateScarlet@Gmail.com',
    'url': 'https://github.com/NateScarlet/graphene-resolver',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Relay compatible connection resolver.  """
import re
import typing

import graphene
from graphql_relay.connection import arrayconnection
import lazy_object_proxy as lazy

from . import resolver
from . import schema as schema_

REGISTRY: typing.Dict[str, typing.Type] = {}


def _get_node_name(node: typing.Union[resolver.Resolver, str, typing.Any]) -> str:
    if isinstance(node, str):
        return node
    if (isinstance(node, type)
            and issubclass(node, resolver.Resolver)):
        node_name = schema_.FieldDefinition.parse(
            node.schema, default={'name': node.__name__}).name
    else:
        node_name = schema_.FieldDefinition.parse(node).name
    return node_name


def build_schema(
        node: typing.Union[resolver.Resolver, str, typing.Any],
        *,
        name: str = None,) -> dict:
    """Build a github-like connection resolver schema.
    see at https://developer.github.com/v4/explorer/

    Args:
        node (typing.Union[resolver.Resolver, str, typing.Any]): Node resolver or schema.
        name (str, optional): Override default connection name,
            required when node name is not defined.

    Returns:
        dict: dict for Resolver schema.
    """

    name = name or f'{_get_node_name(node)}Connection'
    edge_name = f"{re.sub('Connection$', '', name)}Edge"

    return dict(
        name=name,
        description=f"The connection type for {re.sub('Connection$', '', name)}.",
        args=dict(
            after={
                'type': 'String',
                'description': ('Returns the elements in the list '
                                'that come after the specified cursor.')
            },
            before={
                'type': 'String',
                'description': ('Returns the elements in the list '
                                'that come before the specified cursor.')
            },
            first={
                'type': 'Int',
                'description': 'Returns the first _n_ elements from the list.'
            },
            last={
                'type': 'Int',
                'description': 'Returns the last _n_ elements from the list.'
            },
        ),
        type={
            'edges': {
                'type': [{
                    'name': edge_name,
                    'type': {
                        'node': {
                            'type': node,
                            'description': 'The item at the end of the edge.',
                        },
                        'cursor': {
                            'type': 'String!',
                            'description': 'A cursor for use in pagination.'
                        },
                    },
                }],
                'description': 'A list of edges.'
            },
            'nodes': {
                'type': [node],
                'description': 'A list of nodes.'
            },
            'pageInfo': {
                'type': graphene.relay.PageInfo,
                'required': True,
                'description': 'Information to aid in pagination.',
            },
            'totalCount': {
                'type': 'Int!',
                'description': 'Identifies the total count of items in the connection.',
            },
        }
    )


def get_type(
        node: typing.Union[resolver.Resolver, str, typing.Any],
        *,
        name: str = None,
) -> resolver.Resolver:
    """Get connection resolver from registry.
    one will be created with `build_schema` if not found in registry.

    Args:
        node (typing.Union[resolver.Resolver, str, typing.Any]): Node resolver or schema.
        name (str, optional): Override default connection name,
            required when node name is not defined.

    Returns:
        resolver.Resolver: Created connection resolver, same name will returns same resolver.
    """

    name = name or f'{_get_node_name(node)}Connection'

    if name in REGISTRY:
        return REGISTRY[name]

    REGISTRY[name] = type(
        name, (resolver.Resolver,),
        dict(schema=build_schema(node, name=name))
    )

    return REGISTRY[name]


def resolve(
        iterable,
        length: int = None,
        *,
        first: int = None,
        last: int = None,
        after: str = None,
        before: str = None,
        **_,
) -> dict:
    """Resolve iterable to connection

    Args:
        iterable (typing.Iterable): value
        length (int, Optional): defaults to `len(iterable)`,
            iterable length.

    Returns:
        dict: Connection data.
    """
    if length is None:
        length = len(iterable)

    before_offset = arrayconnection.get_offset_with_default(before, length)
    after_offset = arrayconnection.get_offset_with_default(after, -1)
    start_offset = max(after_offset, -1) + 1
    end_offset = min(before_offset, length)
    if isinstance(first, int):
        end_offset = min(end_offset, start_offset + first)
    if isinstance(last, int):
        start_offset = max(start_offset, end_offset - last)

    nodes = lazy.Proxy(lambda: list(iterable[start_offset:end_offset]))
    edges = lazy.Proxy(lambda: [
        dict(
            node=node,
            cursor=arrayconnection.offset_to_cursor(start_offset + i)
        )
        for i, node in enumerate(nodes)
    ])

    return dict(
        nodes=nodes,
        edges=edges,
        pageInfo=lazy.Proxy(lambda: dict(
            start_cursor=edges[0]['cursor'] if edges else None,
            end_cursor=edges[-1]['cursor'] if edges else None,
            has_previous_page=isinstance(
                last, int) and start_offset > (after_offset + 1 if after else 0),
            has_next_page=isinstance(
                first, int) and end_offset < (before_offset if before else length),
        )),
        totalCount=length,
    )



"""Using mongoose-like schema to write apollo-like resolver for graphene.  """

__version__ = '0.1.0'
from .resolver import Resolver
from . import connection, typedef
from .schema import CONFIG_PROCESSOR
